from __future__ import absolute_import, division, print_function, unicode_literals

from .graphics import (save_rgb, view, view_indexed, view_cube, view_nd,
                      get_rgb)
from .spypylab import imshow, ImageView
from .colorscale import ColorScale

